from typing import Any, Protocol

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.base import RepositoryProtocol


class _UpdatedByProvider[PKType, ModelType: GenericBaseModel](Protocol):
    repo: RepositoryProtocol[PKType, ModelType]
    updated_by_attribute: str


class UpdatedByServiceMixin[PKType, ModelType: GenericBaseModel]:
    updated_by_attribute: str = 'updated_by'

    def _apply_updated_by(
        self: _UpdatedByProvider[PKType, ModelType],
        payload: dict[str, Any],
        actor_id: Any | None,
    ) -> dict[str, Any]:
        if actor_id is None:
            return payload
        model = self.repo.model
        if not hasattr(model, self.updated_by_attribute):
            msg = (
                f'updated_by field "{self.updated_by_attribute}" is not defined on '
                f'{model.__name__}.'
            )
            raise ValueError(msg)
        if self.updated_by_attribute not in payload:
            payload[self.updated_by_attribute] = actor_id
        return payload



from collections.abc import Iterable
from typing import Any

from pydantic import BaseModel as PydanticModel
from sqlalchemy.ext.asyncio import AsyncSession

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.types import FilterSpec, OptionSpec
from notora.v2.schemas.base import BaseResponseSchema
from notora.v2.services.mixins.accessors import RepositoryAccessorMixin
from notora.v2.services.mixins.executor import SessionExecutorMixin
from notora.v2.services.mixins.m2m import ManyToManySyncMixin
from notora.v2.services.mixins.payload import PayloadMixin
from notora.v2.services.mixins.serializer import SerializerProtocol
from notora.v2.services.mixins.updated_by import UpdatedByServiceMixin

__all__ = ['UpdateByFilterServiceMixin', 'UpdateServiceMixin']


class UpdateServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    ManyToManySyncMixin[PKType, ModelType],
    UpdatedByServiceMixin[PKType, ModelType],
    PayloadMixin[ModelType],
    SerializerProtocol[ModelType, DetailSchema, ListSchema],
):
    async def update_raw(
        self,
        session: AsyncSession,
        pk: PKType,
        data: PydanticModel | dict[str, Any],
        *,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ModelType:
        payload = self._dump_payload(data, exclude_unset=True)
        payload, relation_payload = self.split_m2m_payload(payload)
        payload = self._apply_updated_by(payload, actor_id)
        query = self.repo.update(pk, payload, options=options)
        entity = await self.execute_for_one(session, query)
        if relation_payload:
            await self.sync_m2m_relations(session, self._extract_pk(entity), relation_payload)
        return entity

    async def update(
        self,
        session: AsyncSession,
        pk: PKType,
        data: PydanticModel | dict[str, Any],
        *,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: type[DetailSchema] | None = None,
    ) -> DetailSchema:
        entity = await self.update_raw(session, pk, data, actor_id=actor_id, options=options)
        return self.serialize_one(entity, schema=schema)


class UpdateByFilterServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    SessionExecutorMixin[PKType, ModelType],
    RepositoryAccessorMixin[PKType, ModelType],
    UpdatedByServiceMixin[PKType, ModelType],
    PayloadMixin[ModelType],
    SerializerProtocol[ModelType, DetailSchema, ListSchema],
):
    async def update_by_raw(
        self,
        session: AsyncSession,
        filters: Iterable[FilterSpec[ModelType]],
        data: PydanticModel | dict[str, Any],
        *,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ModelType:
        payload = self._dump_payload(data, exclude_unset=True)
        payload = self._apply_updated_by(payload, actor_id)
        query = self.repo.update_by(payload, filters=filters, options=options)
        return await self.execute_for_one(session, query)

    async def update_by(
        self,
        session: AsyncSession,
        filters: Iterable[FilterSpec[ModelType]],
        data: PydanticModel | dict[str, Any],
        *,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: type[DetailSchema] | None = None,
    ) -> DetailSchema:
        entity = await self.update_by_raw(
            session,
            filters,
            data,
            actor_id=actor_id,
            options=options,
        )
        return self.serialize_one(entity, schema=schema)



from collections.abc import AsyncIterator, Iterator

import pytest
from sqlalchemy import text
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from testcontainers.postgres import PostgresContainer

from notora.v2.services.config import ServiceConfig
from tests.v2.test_integration.mocks.model import V2Profile, V2Role, V2User, V2UserRole
from tests.v2.test_integration.mocks.repo import (
    V2ProfileRepo,
    V2RoleRepo,
    V2UserRepo,
    V2UserRoleRepo,
)
from tests.v2.test_integration.mocks.schema import (
    V2ProfileResponseSchema,
    V2UserResponseSchema,
)
from tests.v2.test_integration.mocks.service import V2ProfileService, V2UserService


@pytest.fixture(scope='session')
def postgres_db(request) -> Iterator[PostgresContainer]:  # type: ignore[no-untyped-def]
    postgres_version = request.config.getoption('--postgres-version')
    with PostgresContainer(f'postgres:{postgres_version}') as db:
        yield db


@pytest.fixture(scope='session')
async def db_engine(postgres_db: PostgresContainer) -> AsyncIterator[AsyncEngine]:
    url = postgres_db.get_connection_url(driver='asyncpg')
    engine = create_async_engine(url)
    yield engine
    await engine.dispose()


@pytest.fixture(scope='session', autouse=True)
async def init_db(db_engine: AsyncEngine) -> None:
    async with db_engine.begin() as conn:
        await conn.execute(text('CREATE EXTENSION IF NOT EXISTS pgcrypto'))
        await conn.run_sync(V2User.metadata.create_all)
        stmt = text("SELECT c.relname FROM pg_class c WHERE c.relkind = 'S';")
        sequences = (await conn.execute(stmt)).scalars().all()
        for sequence in sequences:
            await conn.execute(text(f'ALTER SEQUENCE {sequence} RESTART;'))


@pytest.fixture(autouse=True)
async def clean_all_tables(db_engine: AsyncEngine) -> None:
    stmt = text("SELECT t.table_name FROM information_schema.tables t WHERE table_schema='public'")
    async with db_engine.begin() as conn:
        tables = (await conn.execute(stmt)).scalars().all()
        tables = [
            t_name
            for t_name in tables
            if not t_name.startswith('pg_')
            and not t_name.startswith('v_')
            and not t_name.startswith('alembic')
        ]
        for table_name in tables:
            await conn.execute(text(f'TRUNCATE TABLE "{table_name}" CASCADE;'))


@pytest.fixture(scope='session')
def session_factory(db_engine: AsyncEngine) -> async_sessionmaker[AsyncSession]:
    return async_sessionmaker(db_engine, expire_on_commit=False)


@pytest.fixture(scope='function')
async def db_session(
    session_factory: async_sessionmaker[AsyncSession],
) -> AsyncIterator[AsyncSession]:
    async with session_factory() as session:
        try:
            yield session
        finally:
            await session.rollback()


@pytest.fixture(scope='function')
def user_repo() -> V2UserRepo:
    return V2UserRepo(V2User)


@pytest.fixture(scope='function')
def role_repo() -> V2RoleRepo:
    return V2RoleRepo(V2Role)


@pytest.fixture(scope='function')
def user_role_repo() -> V2UserRoleRepo:
    return V2UserRoleRepo(V2UserRole)


@pytest.fixture(scope='function')
def profile_repo() -> V2ProfileRepo:
    return V2ProfileRepo(V2Profile)


@pytest.fixture(scope='function')
def user_service(user_repo: V2UserRepo) -> V2UserService:
    config = ServiceConfig(
        detail_schema=V2UserResponseSchema,
        list_schema=V2UserResponseSchema,
    )
    return V2UserService(user_repo, config=config)


@pytest.fixture(scope='function')
def profile_service(profile_repo: V2ProfileRepo) -> V2ProfileService:
    config = ServiceConfig(detail_schema=V2ProfileResponseSchema)
    return V2ProfileService(profile_repo, config=config)



from collections.abc import Callable, Sequence
from dataclasses import dataclass
from enum import StrEnum
from typing import Any

from sqlalchemy import delete, insert, select
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import InstrumentedAttribute

from notora.v2.models.base import GenericBaseModel
from notora.v2.services.mixins.accessors import RepositoryAccessorMixin
from notora.v2.services.mixins.executor import SessionExecutorMixin


class M2MSyncMode(StrEnum):
    REPLACE = 'replace'
    ADD = 'add'
    REMOVE = 'remove'


@dataclass(slots=True)
class ManyToManyRelation[ModelType: GenericBaseModel]:
    payload_field: str
    association_model: type[ModelType]
    left_key: InstrumentedAttribute[Any]
    right_key: InstrumentedAttribute[Any]
    row_factory: Callable[[Any, Any], dict[str, Any]] | None = None

    def build_row(self, owner_id: Any, target_id: Any) -> dict[str, Any]:
        if self.row_factory:
            return self.row_factory(owner_id, target_id)
        return {self.left_key.key: owner_id, self.right_key.key: target_id}


class ManyToManySyncMixin[PKType, ModelType: GenericBaseModel](
    SessionExecutorMixin[PKType, ModelType],
    RepositoryAccessorMixin[PKType, ModelType],
):
    many_to_many_relations: Sequence[ManyToManyRelation[Any]] = ()
    m2m_sync_mode: M2MSyncMode = M2MSyncMode.REPLACE

    def split_m2m_payload(
        self,
        payload: dict[str, Any],
    ) -> tuple[dict[str, Any], dict[str, Sequence[Any]]]:
        if not self.many_to_many_relations:
            return payload, {}
        data = dict(payload)
        relation_payload: dict[str, Sequence[Any]] = {}
        for relation in self.many_to_many_relations:
            if relation.payload_field in data:
                relation_payload[relation.payload_field] = data.pop(relation.payload_field) or ()
        return data, relation_payload

    async def sync_m2m_relations(  # noqa: C901, PLR0912
        self,
        session: AsyncSession,
        owner_id: PKType,
        relation_payload: dict[str, Sequence[Any]],
        *,
        mode: M2MSyncMode | str | None = None,
    ) -> None:
        sync_mode_raw = mode or self.m2m_sync_mode
        if isinstance(sync_mode_raw, M2MSyncMode):
            sync_mode: M2MSyncMode | None = sync_mode_raw
        elif isinstance(sync_mode_raw, str):
            try:
                sync_mode = M2MSyncMode(sync_mode_raw)
            except ValueError:
                sync_mode = None
        else:
            sync_mode = None
        if sync_mode is None:
            msg = 'm2m sync mode must be replace, add, or remove.'
            raise ValueError(msg)
        for relation in self.many_to_many_relations:
            if relation.payload_field not in relation_payload:
                continue
            target_ids = list(dict.fromkeys(relation_payload[relation.payload_field] or ()))
            if sync_mode == M2MSyncMode.REPLACE:
                delete_stmt = delete(relation.association_model).where(
                    relation.left_key == owner_id
                )
                await self.execute(session, delete_stmt)
            elif not target_ids:
                continue
            if not target_ids:
                continue
            if sync_mode == M2MSyncMode.REMOVE:
                delete_stmt = delete(relation.association_model).where(
                    relation.left_key == owner_id,
                    relation.right_key.in_(target_ids),
                )
                await self.execute(session, delete_stmt)
                continue
            if sync_mode == M2MSyncMode.ADD:
                existing: list[Any] = await self.execute_scalars_all(
                    session,
                    select(relation.right_key).where(
                        relation.left_key == owner_id,
                        relation.right_key.in_(target_ids),
                    ),
                )
                existing_set = set(existing)
                target_ids = [
                    target_id for target_id in target_ids if target_id not in existing_set
                ]
                if not target_ids:
                    continue
            rows = [relation.build_row(owner_id, target_id) for target_id in target_ids]
            await self.execute(session, insert(relation.association_model).values(rows))



from collections.abc import Iterable
from typing import Any

from sqlalchemy import Executable
from sqlalchemy.ext.asyncio import AsyncSession

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.params import PaginationParams
from notora.v2.repositories.types import FilterSpec, OptionSpec, OrderSpec
from notora.v2.schemas.base import (
    BaseResponseSchema,
    PaginatedResponseSchema,
    PaginationMetaSchema,
)
from notora.v2.services.mixins.listing import ListingServiceMixin

__all__ = ['PaginationServiceMixin']


class PaginationServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    ListingServiceMixin[PKType, ModelType, DetailSchema, ListSchema],
):
    async def paginate(
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        limit: int = 20,
        offset: int = 0,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        base_query: Any | None = None,
        schema: type[ListSchema] | None = None,
    ) -> 'PaginatedResponseSchema[ListSchema]':
        data = await self.list_raw(
            session,
            filters=filters,
            limit=limit,
            offset=offset,
            ordering=ordering,
            options=options,
            base_query=base_query,
        )
        serialized = self.serialize_many(data, schema=schema)
        total_query = self.repo.count(filters=filters)
        total: int = await self.execute_scalar_one(session, total_query)
        meta = PaginationMetaSchema.calculate(total=total, limit=limit, offset=offset)
        return PaginatedResponseSchema(meta=meta, data=serialized)

    async def paginate_from_queries(
        self,
        session: AsyncSession,
        *,
        data_query: Executable,
        count_query: Executable,
        limit: int,
        offset: int,
        schema: type[ListSchema] | None = None,
    ) -> 'PaginatedResponseSchema[ListSchema]':
        data: list[ModelType] = await self.execute_scalars_all(session, data_query)
        serialized = self.serialize_many(data, schema=schema)
        total: int = await self.execute_scalar_one(session, count_query)
        meta = PaginationMetaSchema.calculate(total=total, limit=limit, offset=offset)
        return PaginatedResponseSchema(meta=meta, data=serialized)

    async def paginate_params(
        self,
        session: AsyncSession,
        params: PaginationParams[ModelType],
        *,
        schema: type[ListSchema] | None = None,
    ) -> 'PaginatedResponseSchema[ListSchema]':
        return await self.paginate(
            session,
            filters=params.filters,
            limit=params.limit,
            offset=params.offset,
            ordering=params.ordering,
            options=params.options,
            base_query=params.base_query,
            schema=schema,
        )



from uuid import UUID

import pytest
from sqlalchemy import select
from sqlalchemy.exc import NoResultFound
from sqlalchemy.ext.asyncio import AsyncSession

from notora.v1.schemas.base import Filter
from tests.v1.test_integration.mocks.model import MockModel, MockModelFactory
from tests.v1.test_integration.mocks.repo import MockRepo


async def _setup(db_session: AsyncSession, count: int) -> list[MockModel]:
    objs = [MockModelFactory.build() for _ in range(count)]
    db_session.add_all(objs)
    await db_session.commit()
    return objs


async def _select_one(db_session: AsyncSession, pk: UUID) -> MockModel:
    query = select(MockModel).where(MockModel.id == pk)
    return (await db_session.scalars(query)).one()


async def test_select(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    objs_count = 3
    objs = await _setup(db_session, objs_count)

    query = mock_repo.select()

    res = (await db_session.scalars(query)).all()
    assert len(res) == objs_count
    assert set(objs) == set(res)


async def test_create(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    new_data = {'name': 'test'}

    query = mock_repo.create(new_data)

    res = (await db_session.scalars(query)).one()
    assert res.id is not None


async def test_retrieve(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    objs = await _setup(db_session, 3)
    target_obj = objs[0]

    query = mock_repo.retrieve(target_obj.id)

    res = (await db_session.scalars(query)).one_or_none()
    assert res is not None
    assert res.id == target_obj.id


async def test_retrieve_by(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    objs = await _setup(db_session, 3)
    target_obj = objs[0]

    query = mock_repo.retrieve_by(filters=[Filter(field='id', op='=', value=target_obj.id)])

    res = (await db_session.scalars(query)).one_or_none()
    assert res is not None
    assert res.id == target_obj.id


async def test_update(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    objs = await _setup(db_session, 1)
    target_obj = objs[0]
    new_value = 'Updated Value'

    query = mock_repo.update(target_obj.id, {'name': new_value})
    result = await db_session.execute(query)
    result.scalars().all()
    await db_session.commit()

    updated_obj = await _select_one(db_session, target_obj.id)
    assert updated_obj.name == new_value


async def test_update_by(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    objs = await _setup(db_session, 3)
    target_obj = objs[0]
    new_value = 'Updated Value'

    query = mock_repo.update_by(
        {'name': new_value}, filters=[Filter(field='id', op='=', value=target_obj.id)]
    )
    result = await db_session.execute(query)
    result.scalars().all()
    await db_session.commit()

    updated_obj = await _select_one(db_session, target_obj.id)
    assert updated_obj.name == new_value


async def test_hard_delete(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    objs = await _setup(db_session, 1)
    target_obj = objs[0]

    query = mock_repo.hard_delete(target_obj.id)
    result = await db_session.execute(query)
    result.scalars().all()
    await db_session.commit()

    with pytest.raises(NoResultFound):
        await _select_one(db_session, target_obj.id)


async def test_hard_delete_by(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    objs = await _setup(db_session, 3)
    ids_to_delete = [obj.id for obj in objs[:2]]  # Delete first two

    query = mock_repo.hard_delete_by(filters=[Filter(field='id', op='in', value=ids_to_delete)])
    result = await db_session.execute(query)
    result.scalars().all()
    await db_session.commit()

    # Ensure the deleted objects are no longer in the database
    for obj_id in ids_to_delete:
        with pytest.raises(NoResultFound):
            await _select_one(db_session, obj_id)

    # Verify the remaining object still exists
    assert await _select_one(db_session, objs[2].id)


async def test_number_of_records(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    objs_count = 5
    await _setup(db_session, objs_count)

    query = mock_repo.number_of_records()

    total_count = (await db_session.scalars(query)).one()
    assert total_count == objs_count


async def test_soft_delete(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    objs = await _setup(db_session, 1)
    obj_to_delete = objs[0]

    query = mock_repo.soft_delete(obj_to_delete.id)
    result = await db_session.execute(query)
    result.scalars().all()
    await db_session.commit()

    refreshed_obj = await db_session.get(MockModel, obj_to_delete.id)
    assert refreshed_obj
    assert refreshed_obj.deleted_at is not None


async def test_soft_delete_by(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    objs = await _setup(db_session, 3)
    ids_to_delete = [obj.id for obj in objs[:2]]  # Delete first two

    query = mock_repo.soft_delete_by(filters=[Filter(field='id', op='in', value=ids_to_delete)])
    result = await db_session.execute(query)
    result.scalars().all()
    await db_session.commit()

    # Verify that the first two are marked as deleted
    for obj_id in ids_to_delete:
        obj = await _select_one(db_session, obj_id)
        assert obj.deleted_at is not None
    # Verify the third is still not deleted
    remaining_obj = await _select_one(db_session, objs[2].id)
    assert remaining_obj
    assert remaining_obj.deleted_at is None


async def test_select_excludes_deleted(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    objs = await _setup(db_session, 2)

    # Soft delete one object
    obj_to_delete = objs[0]
    query = mock_repo.soft_delete(obj_to_delete.id)
    result = await db_session.execute(query)
    result.scalars().all()
    await db_session.commit()

    # Ensure only non-deleted objects are retrieved
    query = mock_repo.select()
    res = (await db_session.scalars(query)).all()
    assert len(res) == 1
    assert res[0].id != obj_to_delete.id


async def test_soft_delete_is_not_hard(db_session: AsyncSession, mock_repo: MockRepo) -> None:
    objs_count = 3
    objs = await _setup(db_session, objs_count)

    # Soft delete one object
    obj_to_delete = objs[0]
    query = mock_repo.soft_delete(obj_to_delete.id)
    result = await db_session.execute(query)
    result.scalars().all()
    await db_session.commit()

    # Check that all objects (including deleted) can be retrieved
    res = (await db_session.scalars(select(mock_repo.model))).all()
    assert len(res) == objs_count
    assert obj_to_delete in res



import pytest
from pydantic import ValidationError
from sqlalchemy import Integer, String
from sqlalchemy.orm import InstrumentedAttribute, Mapped, mapped_column

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories import FilterField, QueryInput, SortField, build_query_params
from notora.v2.repositories.types import DEFAULT_LIMIT


class Widget(GenericBaseModel):
    __tablename__ = 'widget_query'

    name: Mapped[str] = mapped_column(String)
    count: Mapped[int] = mapped_column(Integer)


EXPECTED_LIMIT = 10
EXPECTED_OFFSET = 5
EXPECTED_FILTERS = 2
EXPECTED_ORDERING = 1


def _name_column(model: type[Widget]) -> InstrumentedAttribute[str]:
    return model.name


def _count_column(model: type[Widget]) -> InstrumentedAttribute[int]:
    return model.count


def test_build_query_params_builds_filters_and_sort() -> None:
    query = QueryInput(
        filter=['name:eq:alpha', 'count:gte:2'],
        sort=['-name'],
        limit=EXPECTED_LIMIT,
        offset=EXPECTED_OFFSET,
    )
    filter_fields: dict[str, FilterField[Widget]] = {
        'name': FilterField(resolver=_name_column, value_type=str),
        'count': FilterField(resolver=_count_column, value_type=int),
    }
    sort_fields: dict[str, SortField[Widget]] = {'name': SortField(resolver=_name_column)}

    params = build_query_params(
        query,
        model=Widget,
        filter_fields=filter_fields,
        sort_fields=sort_fields,
    )

    assert params.limit == EXPECTED_LIMIT
    assert params.offset == EXPECTED_OFFSET
    assert params.filters is not None
    assert params.ordering is not None
    assert len(list(params.filters)) == EXPECTED_FILTERS
    assert len(list(params.ordering)) == EXPECTED_ORDERING


def test_build_query_params_rejects_unknown_filter_field() -> None:
    query = QueryInput(filter=['unknown:eq:1'])
    filter_fields: dict[str, FilterField[Widget]] = {
        'name': FilterField(resolver=_name_column, value_type=str),
    }

    with pytest.raises(ValueError, match='Unsupported filter field'):
        build_query_params(query, model=Widget, filter_fields=filter_fields)


def test_build_query_params_accepts_direct_attributes() -> None:
    query = QueryInput(filter=['name:eq:alpha'], sort=['+count'])
    filter_fields: dict[str, FilterField[Widget]] = {
        'name': FilterField(resolver=Widget.name, value_type=str),
    }
    sort_fields: dict[str, SortField[Widget]] = {
        'count': SortField(resolver=Widget.count),
    }

    params = build_query_params(
        query,
        model=Widget,
        filter_fields=filter_fields,
        sort_fields=sort_fields,
    )

    assert params.limit is DEFAULT_LIMIT
    assert params.filters is not None
    assert params.ordering is not None


def test_build_query_params_parses_in_and_isnull() -> None:
    query = QueryInput(filter=['count:in:1,2', 'name:isnull'])
    filter_fields: dict[str, FilterField[Widget]] = {
        'name': FilterField(resolver=_name_column, value_type=str),
        'count': FilterField(resolver=_count_column, value_type=int),
    }

    params = build_query_params(query, model=Widget, filter_fields=filter_fields)
    clauses = list(params.filters or ())

    assert ' IN ' in str(clauses[0])
    assert ' IS NULL' in str(clauses[1])


def test_build_query_params_rejects_disallowed_operator() -> None:
    query = QueryInput(filter=['name:ilike:alpha'])
    filter_fields: dict[str, FilterField[Widget]] = {
        'name': FilterField(resolver=_name_column, operators=frozenset({'eq'})),
    }

    with pytest.raises(ValueError, match='Operator'):
        build_query_params(query, model=Widget, filter_fields=filter_fields)


def test_query_input_rejects_non_positive_limit() -> None:
    with pytest.raises(ValidationError, match='limit must be a positive integer'):
        QueryInput(limit=0)



from dataclasses import dataclass

from notora.v2.schemas.base import BaseResponseSchema


@dataclass(slots=True)
class ServiceConfig[
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
]:
    detail_schema: type[DetailSchema] | None = None
    list_schema: type[ListSchema] | None = None


type ServiceConfigD[DetailSchema: BaseResponseSchema] = ServiceConfig[DetailSchema, DetailSchema]



from uuid import uuid4

from sqlalchemy.ext.asyncio import AsyncSession

from tests.v2.test_integration.mocks.model import V2User
from tests.v2.test_integration.mocks.repo import V2UserRepo


async def _create_user(
    session: AsyncSession,
    repo: V2UserRepo,
    *,
    email: str,
    name: str,
    is_active: bool = True,
) -> V2User:
    payload = {
        'id': uuid4(),
        'email': email,
        'name': name,
        'is_active': is_active,
    }
    user = (await session.scalars(repo.create(payload))).one()
    await session.commit()
    return user


async def _seed_users(session: AsyncSession) -> list[V2User]:
    users = [
        V2User(id=uuid4(), email='a@ex.com', name='Alice', is_active=True),
        V2User(id=uuid4(), email='b@ex.com', name='Bob', is_active=False),
        V2User(id=uuid4(), email='c@ex.com', name='Cara', is_active=True),
    ]
    session.add_all(users)
    await session.commit()
    return users


async def test_repo_create_retrieve_and_count(
    db_session: AsyncSession,
    user_repo: V2UserRepo,
) -> None:
    user = await _create_user(
        db_session,
        user_repo,
        email='one@ex.com',
        name='One',
    )

    query = user_repo.retrieve(user.id)
    retrieved = (await db_session.scalars(query)).one()
    assert retrieved.id == user.id

    total = (await db_session.execute(user_repo.count())).scalar_one()
    assert total == 1


async def test_repo_list_filters_and_ordering(
    db_session: AsyncSession,
    user_repo: V2UserRepo,
) -> None:
    await _seed_users(db_session)

    query = user_repo.list(
        filters=[V2User.is_active.is_(True)],
        ordering=[V2User.email.desc()],
        limit=None,
    )
    users = (await db_session.scalars(query)).all()

    assert [user.email for user in users] == ['c@ex.com', 'a@ex.com']


async def test_repo_update_and_delete(
    db_session: AsyncSession,
    user_repo: V2UserRepo,
) -> None:
    user = await _create_user(
        db_session,
        user_repo,
        email='update@ex.com',
        name='Before',
    )

    await db_session.execute(user_repo.update(user.id, {'name': 'After'}))
    await db_session.commit()

    refreshed = await db_session.get(V2User, user.id)
    assert refreshed is not None
    assert refreshed.name == 'After'

    await db_session.execute(user_repo.delete(user.id))
    await db_session.commit()

    assert await db_session.get(V2User, user.id) is None


async def test_repo_delete_by(
    db_session: AsyncSession,
    user_repo: V2UserRepo,
) -> None:
    users = await _seed_users(db_session)
    delete_query = user_repo.delete_by(
        filters=[V2User.email.in_([users[0].email, users[1].email])]
    )
    await db_session.execute(delete_query)
    await db_session.commit()

    remaining = (await db_session.scalars(user_repo.list(limit=None))).all()
    assert [user.email for user in remaining] == [users[2].email]


async def test_repo_soft_delete(
    db_session: AsyncSession,
    user_repo: V2UserRepo,
) -> None:
    user = await _create_user(
        db_session,
        user_repo,
        email='soft@ex.com',
        name='Soft',
    )

    await db_session.execute(user_repo.soft_delete(user.id))
    await db_session.commit()

    refreshed = await db_session.get(V2User, user.id)
    assert refreshed is not None
    assert refreshed.deleted_at is not None


async def test_repo_create_or_skip(
    db_session: AsyncSession,
    user_repo: V2UserRepo,
) -> None:
    payload = {'id': uuid4(), 'email': 'skip@ex.com', 'name': 'Skip', 'is_active': True}

    create_query = user_repo.create_or_skip(payload, conflict_columns=[V2User.email])
    created = (await db_session.scalars(create_query)).one_or_none()
    await db_session.commit()
    assert created is not None

    duplicate_payload = {**payload, 'id': uuid4()}
    duplicate_query = user_repo.create_or_skip(
        duplicate_payload,
        conflict_columns=[V2User.email],
    )
    duplicate = (await db_session.scalars(duplicate_query)).one_or_none()
    await db_session.commit()
    assert duplicate is None


async def test_repo_upsert_updates_existing(
    db_session: AsyncSession,
    user_repo: V2UserRepo,
) -> None:
    payload = {'id': uuid4(), 'email': 'upsert@ex.com', 'name': 'Before', 'is_active': True}
    user = (await db_session.scalars(user_repo.create(payload))).one()
    await db_session.commit()

    upsert_payload = {
        'id': uuid4(),
        'email': user.email,
        'name': 'After',
        'is_active': True,
    }
    upsert_query = user_repo.upsert(
        upsert_payload,
        conflict_columns=[V2User.email],
        update_only=['name'],
    )
    updated = (await db_session.scalars(upsert_query)).one()
    await db_session.commit()

    assert updated.id == user.id
    refreshed = await db_session.get(V2User, user.id)
    assert refreshed is not None
    await db_session.refresh(refreshed)
    assert refreshed.name == 'After'



from notora.v2.repositories.mixins.query import (
    CountableMixin,
    FilterableMixin,
    ListableMixin,
    LoadOptionsMixin,
    OrderableMixin,
    PrimaryKeyMixin,
    RetrievableMixin,
    SelectableMixin,
)
from notora.v2.repositories.mixins.write import (
    CreateMixin,
    CreateOrSkipMixin,
    DeleteMixin,
    SoftDeleteMixin,
    UpdateMixin,
    UpsertMixin,
)

__all__ = [
    'CountableMixin',
    'CreateMixin',
    'CreateOrSkipMixin',
    'DeleteMixin',
    'FilterableMixin',
    'ListableMixin',
    'LoadOptionsMixin',
    'OrderableMixin',
    'PrimaryKeyMixin',
    'RetrievableMixin',
    'SelectableMixin',
    'SoftDeleteMixin',
    'UpdateMixin',
    'UpsertMixin',
]



from uuid import UUID

from notora.v2.repositories.base import Repository, SoftDeleteRepository
from tests.v2.test_integration.mocks.model import V2Profile, V2Role, V2User, V2UserRole


class V2UserRepo(SoftDeleteRepository[UUID, V2User]): ...


class V2RoleRepo(Repository[UUID, V2Role]): ...


class V2UserRoleRepo(Repository[UUID, V2UserRole]): ...


class V2ProfileRepo(Repository[UUID, V2Profile]): ...



from tests.v2.test_integration.mocks.model import (
    V2Profile,
    V2Role,
    V2User,
    V2UserRole,
)
from tests.v2.test_integration.mocks.repo import (
    V2ProfileRepo,
    V2RoleRepo,
    V2UserRepo,
    V2UserRoleRepo,
)
from tests.v2.test_integration.mocks.schema import (
    V2ProfileResponseSchema,
    V2UserCreateSchema,
    V2UserResponseSchema,
)
from tests.v2.test_integration.mocks.service import V2ProfileService, V2UserService

__all__ = [
    'V2Profile',
    'V2ProfileRepo',
    'V2ProfileResponseSchema',
    'V2ProfileService',
    'V2Role',
    'V2RoleRepo',
    'V2User',
    'V2UserCreateSchema',
    'V2UserRepo',
    'V2UserResponseSchema',
    'V2UserRole',
    'V2UserRoleRepo',
    'V2UserService',
]



import operator
from collections.abc import Callable, Iterable, Mapping
from dataclasses import dataclass
from typing import Any, Literal, cast

from pydantic import BaseModel, Field, TypeAdapter, field_validator
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.sql import ColumnElement

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.params import QueryParams
from notora.v2.repositories.types import DEFAULT_LIMIT, DefaultLimit, OrderClause

type FilterOperator = Literal['eq', 'ne', 'lt', 'lte', 'gt', 'gte', 'in', 'ilike', 'isnull']
type SortDirection = Literal['asc', 'desc']
type FilterResolver[ModelType: GenericBaseModel] = (
    InstrumentedAttribute[Any]
    | ColumnElement[Any]
    | Callable[[type[ModelType]], InstrumentedAttribute[Any] | ColumnElement[Any]]
)
type SortResolver[ModelType: GenericBaseModel] = (
    InstrumentedAttribute[Any]
    | ColumnElement[Any]
    | Callable[[type[ModelType]], InstrumentedAttribute[Any] | ColumnElement[Any]]
)
type FilterPredicate[ModelType: GenericBaseModel] = Callable[
    [type[ModelType], FilterOperator, Any], ColumnElement[bool]
]

DEFAULT_OPERATORS: frozenset[FilterOperator] = frozenset({
    'eq',
    'ne',
    'lt',
    'lte',
    'gt',
    'gte',
    'in',
    'ilike',
    'isnull',
})
FILTER_PARTS_MIN = 2
FILTER_PARTS_WITH_VALUE = 3


class QueryInput(BaseModel):
    filter: list[str] = Field(default_factory=list)
    sort: list[str] = Field(default_factory=list)
    limit: int | None = None
    offset: int = 0

    @field_validator('limit')
    @classmethod
    def _validate_limit(cls, value: int | None) -> int | None:
        if value is None:
            return value
        if value <= 0:
            msg = 'limit must be a positive integer.'
            raise ValueError(msg)
        return value

    @field_validator('offset')
    @classmethod
    def _validate_offset(cls, value: int) -> int:
        if value < 0:
            msg = 'offset must be zero or a positive integer.'
            raise ValueError(msg)
        return value


@dataclass(frozen=True, slots=True)
class FilterToken:
    field: str
    operator: FilterOperator
    raw_value: str | None


@dataclass(frozen=True, slots=True)
class SortToken:
    field: str
    direction: SortDirection


@dataclass(frozen=True, slots=True)
class FilterField[ModelType: GenericBaseModel]:
    resolver: FilterResolver[ModelType] | None = None
    predicate: FilterPredicate[ModelType] | None = None
    operators: frozenset[FilterOperator] = DEFAULT_OPERATORS
    value_type: type[Any] | TypeAdapter[Any] | None = None


@dataclass(frozen=True, slots=True)
class SortField[ModelType: GenericBaseModel]:
    resolver: SortResolver[ModelType]


def parse_filter_token(raw: str) -> FilterToken:
    parts = raw.split(':', 2)
    if len(parts) < FILTER_PARTS_MIN:
        msg = 'filter must be in "field:op:value" format.'
        raise ValueError(msg)
    field_name = parts[0].strip()
    op = parts[1].strip()
    raw_value = parts[2] if len(parts) == FILTER_PARTS_WITH_VALUE else None
    if not field_name:
        msg = 'filter field name cannot be empty.'
        raise ValueError(msg)
    if op not in DEFAULT_OPERATORS:
        msg = f'Unsupported filter operator "{op}".'
        raise ValueError(msg)
    operator = cast(FilterOperator, op)
    if raw_value is not None:
        raw_value = raw_value.strip()
        if not raw_value:
            raw_value = None
    return FilterToken(field=field_name, operator=operator, raw_value=raw_value)


def parse_sort_token(raw: str) -> SortToken:
    value = raw.strip()
    if not value:
        msg = 'sort field name cannot be empty.'
        raise ValueError(msg)
    direction: SortDirection
    if value[0] == '-':
        direction = 'desc'
        field_name = value[1:]
    elif value[0] == '+':
        direction = 'asc'
        field_name = value[1:]
    else:
        direction = 'asc'
        field_name = value
    field_name = field_name.strip()
    if not field_name:
        msg = 'sort field name cannot be empty.'
        raise ValueError(msg)
    return SortToken(field=field_name, direction=direction)


def _resolve_resolver[ModelType: GenericBaseModel](
    resolver: FilterResolver[ModelType] | SortResolver[ModelType],
    model: type[ModelType],
) -> InstrumentedAttribute[Any] | ColumnElement[Any]:
    return resolver(model) if callable(resolver) else resolver


def _parse_scalar_value(field: FilterField[Any], raw_value: str) -> Any:
    if field.value_type is None:
        return raw_value
    adapter = (
        field.value_type
        if isinstance(field.value_type, TypeAdapter)
        else TypeAdapter(field.value_type)
    )
    return adapter.validate_python(raw_value)


def _parse_filter_value(token: FilterToken, field: FilterField[Any]) -> Any:
    if token.operator == 'isnull':
        if token.raw_value is None:
            return True
        return TypeAdapter(bool).validate_python(token.raw_value)
    if token.raw_value is None:
        msg = f'Filter "{token.field}:{token.operator}" requires a value.'
        raise ValueError(msg)
    if token.operator == 'in':
        items = [item for item in token.raw_value.split(',') if item]
        if not items:
            msg = f'Filter "{token.field}:{token.operator}" requires a value.'
            raise ValueError(msg)
        return [_parse_scalar_value(field, item) for item in items]
    return _parse_scalar_value(field, token.raw_value)


def _apply_operator(
    column: ColumnElement[Any] | InstrumentedAttribute[Any],
    op: FilterOperator,
    value: Any,
) -> ColumnElement[bool]:
    operators: dict[
        FilterOperator,
        Callable[[ColumnElement[Any] | InstrumentedAttribute[Any], Any], ColumnElement[bool]],
    ] = {
        'eq': operator.eq,
        'ne': operator.ne,
        'lt': operator.lt,
        'lte': operator.le,
        'gt': operator.gt,
        'gte': operator.ge,
        'in': lambda col, val: col.in_(val),
        'ilike': lambda col, val: col.ilike(val),
        'isnull': lambda col, val: col.is_(None) if val else col.is_not(None),
    }
    handler = operators.get(op)
    if handler is None:
        msg = f'Unsupported filter operator "{op}".'
        raise ValueError(msg)
    return handler(column, value)


def build_filter_clauses[ModelType: GenericBaseModel](
    tokens: Iterable[FilterToken],
    *,
    model: type[ModelType],
    fields: Mapping[str, FilterField[ModelType]],
) -> list[ColumnElement[bool]]:
    clauses: list[ColumnElement[bool]] = []
    for token in tokens:
        field = fields.get(token.field)
        if field is None:
            msg = f'Unsupported filter field "{token.field}".'
            raise ValueError(msg)
        if token.operator not in field.operators:
            msg = f'Operator "{token.operator}" is not allowed for "{token.field}".'
            raise ValueError(msg)
        value = _parse_filter_value(token, field)
        if field.predicate is not None:
            clause = field.predicate(model, token.operator, value)
        elif field.resolver is not None:
            column = _resolve_resolver(field.resolver, model)
            clause = _apply_operator(column, token.operator, value)
        else:
            msg = f'Filter field "{token.field}" requires resolver or predicate.'
            raise ValueError(msg)
        clauses.append(clause)
    return clauses


def build_sort_clauses[ModelType: GenericBaseModel](
    tokens: Iterable[SortToken],
    *,
    model: type[ModelType],
    fields: Mapping[str, SortField[ModelType]],
) -> list[OrderClause]:
    clauses: list[OrderClause] = []
    for token in tokens:
        field = fields.get(token.field)
        if field is None:
            msg = f'Unsupported sort field "{token.field}".'
            raise ValueError(msg)
        column = _resolve_resolver(field.resolver, model)
        if token.direction == 'desc':
            clauses.append(column.desc())
        else:
            clauses.append(column.asc())
    return clauses


def build_query_params[ModelType: GenericBaseModel](
    query: QueryInput,
    *,
    model: type[ModelType],
    filter_fields: Mapping[str, FilterField[ModelType]] | None = None,
    sort_fields: Mapping[str, SortField[ModelType]] | None = None,
    base_query: Any | None = None,
) -> QueryParams[ModelType]:
    filter_fields = filter_fields or {}
    sort_fields = sort_fields or {}
    if query.filter and not filter_fields:
        msg = 'Filter fields mapping is required when filters are provided.'
        raise ValueError(msg)
    if query.sort and not sort_fields:
        msg = 'Sort fields mapping is required when sorting is provided.'
        raise ValueError(msg)
    filter_tokens = [parse_filter_token(raw) for raw in query.filter]
    sort_tokens = [parse_sort_token(raw) for raw in query.sort]
    filters = (
        build_filter_clauses(filter_tokens, model=model, fields=filter_fields)
        if filter_tokens
        else None
    )
    ordering = (
        build_sort_clauses(sort_tokens, model=model, fields=sort_fields) if sort_tokens else None
    )
    limit: int | DefaultLimit | None = DEFAULT_LIMIT if query.limit is None else query.limit
    return QueryParams(
        filters=filters,
        ordering=ordering,
        limit=limit,
        offset=query.offset,
        base_query=base_query,
    )


def make_query_params_dependency[ModelType: GenericBaseModel](
    *,
    model: type[ModelType],
    filter_fields: Mapping[str, FilterField[ModelType]] | None = None,
    sort_fields: Mapping[str, SortField[ModelType]] | None = None,
    base_query: Any | None = None,
) -> Callable[..., QueryParams[ModelType]]:
    try:
        fastapi = __import__('fastapi')
    except Exception as exc:  # pragma: no cover - only used in FastAPI apps
        msg = 'fastapi is required to use make_query_params_dependency.'
        raise RuntimeError(msg) from exc

    depends = fastapi.Depends()

    def _dependency(query: QueryInput = depends) -> QueryParams[ModelType]:
        return build_query_params(
            query,
            model=model,
            filter_fields=filter_fields,
            sort_fields=sort_fields,
            base_query=base_query,
        )

    return _dependency



class FKNotFoundError(Exception):
    def __init__(self, message: str, fk_name: str, table_name: str) -> None:
        super().__init__(message)
        self.fk_name = fk_name
        self.table_name = table_name


class AlreadyExistsError(Exception):
    def __init__(
        self,
        message: str | None = None,
        constraint_name: str | None = None,
    ) -> None:
        super().__init__(message or 'Entity already exists.')
        self.constraint_name = constraint_name


class NotFoundError[FKType](Exception):
    def __init__(self, *args: object, entity_id: FKType | None = None) -> None:
        super().__init__(*args)
        self.entity_id = entity_id



from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.base import SoftDeleteRepository
from notora.v2.repositories.config import RepoConfig
from notora.v2.repositories.factory import AnyRepository, RepositoryType, build_repository
from notora.v2.schemas.base import BaseResponseSchema
from notora.v2.services.base import RepositoryService, SoftDeleteRepositoryService
from notora.v2.services.config import ServiceConfig

type AnyService[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
] = (
    RepositoryService[PKType, ModelType, DetailSchema, ListSchema]
    | SoftDeleteRepositoryService[PKType, ModelType, DetailSchema, ListSchema]
)

type ServiceType[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
] = (
    type[RepositoryService[PKType, ModelType, DetailSchema, ListSchema]]
    | type[SoftDeleteRepositoryService[PKType, ModelType, DetailSchema, ListSchema]]
)


def build_service[  # noqa: C901
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    model: type[ModelType],
    *,
    repo: AnyRepository[PKType, ModelType] | None = None,
    repo_config: RepoConfig[ModelType] | None = None,
    service_config: ServiceConfig[DetailSchema, ListSchema] | None = None,
    soft_delete: bool = False,
    repo_cls: RepositoryType[PKType, ModelType] | None = None,
    service_cls: ServiceType[PKType, ModelType, DetailSchema, ListSchema] | None = None,
) -> AnyService[PKType, ModelType, DetailSchema, ListSchema]:
    """Create a repository + service pair with optional config overrides."""
    if repo is None:
        repo = build_repository(
            model,
            config=repo_config,
            soft_delete=soft_delete,
            repo_cls=repo_cls,
        )
    if service_cls is None:
        if soft_delete or isinstance(repo, SoftDeleteRepository):
            if not isinstance(repo, SoftDeleteRepository):
                msg = 'Soft-delete service requires a soft-delete repository.'
                raise TypeError(msg)
            return SoftDeleteRepositoryService(repo, config=service_config)
        return RepositoryService(repo, config=service_config)

    if issubclass(service_cls, SoftDeleteRepositoryService):
        if not isinstance(repo, SoftDeleteRepository):
            msg = 'Soft-delete service requires a soft-delete repository.'
            raise TypeError(msg)
        return service_cls(repo, config=service_config)
    return service_cls(repo, config=service_config)


def build_service_for_repo[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    repo: AnyRepository[PKType, ModelType],
    *,
    service_config: ServiceConfig[DetailSchema, ListSchema] | None = None,
    service_cls: ServiceType[PKType, ModelType, DetailSchema, ListSchema] | None = None,
) -> AnyService[PKType, ModelType, DetailSchema, ListSchema]:
    """Create a service for an existing repository."""
    if service_cls is None:
        if isinstance(repo, SoftDeleteRepository):
            return SoftDeleteRepositoryService(repo, config=service_config)
        return RepositoryService(repo, config=service_config)

    if issubclass(service_cls, SoftDeleteRepositoryService):
        if not isinstance(repo, SoftDeleteRepository):
            msg = 'Soft-delete service requires a soft-delete repository.'
            raise TypeError(msg)
        return service_cls(repo, config=service_config)
    return service_cls(repo, config=service_config)



"""Service mixins and helpers for v2."""

from notora.v2.services.base import (
    RepositoryService,
    RepositoryServiceD,
    SoftDeleteRepositoryService,
    SoftDeleteRepositoryServiceD,
)
from notora.v2.services.config import ServiceConfig, ServiceConfigD
from notora.v2.services.factory import build_service, build_service_for_repo
from notora.v2.services.mixins import (
    CreateOrSkipServiceMixin,
    CreateServiceMixin,
    DeleteServiceMixin,
    ListingServiceMixin,
    M2MSyncMode,
    ManyToManyRelation,
    ManyToManySyncMixin,
    PaginationServiceMixin,
    PayloadMixin,
    RepositoryAccessorMixin,
    RetrievalServiceMixin,
    SerializerMixin,
    SerializerProtocol,
    SessionExecutorMixin,
    SoftDeleteServiceMixin,
    UpdateByFilterServiceMixin,
    UpdatedByServiceMixin,
    UpdateServiceMixin,
    UpsertServiceMixin,
)

__all__ = [
    'CreateOrSkipServiceMixin',
    'CreateServiceMixin',
    'DeleteServiceMixin',
    'ListingServiceMixin',
    'M2MSyncMode',
    'ManyToManyRelation',
    'ManyToManySyncMixin',
    'PaginationServiceMixin',
    'PayloadMixin',
    'RepositoryAccessorMixin',
    'RepositoryService',
    'RepositoryServiceD',
    'RetrievalServiceMixin',
    'SerializerMixin',
    'SerializerProtocol',
    'ServiceConfig',
    'ServiceConfigD',
    'SessionExecutorMixin',
    'SoftDeleteRepositoryService',
    'SoftDeleteRepositoryServiceD',
    'SoftDeleteServiceMixin',
    'UpdateByFilterServiceMixin',
    'UpdateServiceMixin',
    'UpdatedByServiceMixin',
    'UpsertServiceMixin',
    'build_service',
    'build_service_for_repo',
]



from collections.abc import Iterable

from sqlalchemy.ext.asyncio import AsyncSession

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.types import FilterSpec, OptionSpec, OrderSpec
from notora.v2.schemas.base import BaseResponseSchema
from notora.v2.services.mixins.accessors import RepositoryAccessorMixin
from notora.v2.services.mixins.executor import SessionExecutorMixin
from notora.v2.services.mixins.serializer import SerializerProtocol

__all__ = ['RetrievalServiceMixin']


class RetrievalServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    SessionExecutorMixin[PKType, ModelType],
    RepositoryAccessorMixin[PKType, ModelType],
    SerializerProtocol[ModelType, DetailSchema, ListSchema],
):
    async def retrieve_raw(
        self,
        session: AsyncSession,
        pk: PKType,
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ModelType:
        query = self.repo.retrieve(pk, options=options)
        return await self.execute_for_one(session, query)

    async def retrieve_one_by_raw(
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ModelType:
        query = self.repo.retrieve_one_by(filters=filters, ordering=ordering, options=options)
        return await self.execute_for_one(session, query)

    async def retrieve(
        self,
        session: AsyncSession,
        pk: PKType,
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: type[DetailSchema] | None = None,
    ) -> DetailSchema:
        entity = await self.retrieve_raw(session, pk, options=options)
        return self.serialize_one(entity, schema=schema)

    async def retrieve_one_by(
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: type[DetailSchema] | None = None,
    ) -> DetailSchema:
        entity = await self.retrieve_one_by_raw(
            session,
            filters=filters,
            ordering=ordering,
            options=options,
        )
        return self.serialize_one(entity, schema=schema)



from typing import cast

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.base import RepositoryProtocol


class RepositoryAccessorMixin[PKType, ModelType: GenericBaseModel]:
    repo: RepositoryProtocol[PKType, ModelType]

    def _extract_pk(self, entity: ModelType) -> PKType:
        pk_attr = getattr(self.repo, 'pk_attribute', 'id')
        return cast(PKType, getattr(entity, pk_attr))



from uuid import UUID

from notora.v1.schemas.base import BaseRequestSchema, BaseResponseSchema


class MockModelResponseSchema(BaseResponseSchema):
    id: UUID
    name: str


class CreateMockModelRequestSchema(BaseRequestSchema):
    name: str


class UpdateMockModelRequestSchema(CreateMockModelRequestSchema): ...



import re
from typing import Any, ClassVar, cast

from sqlalchemy import exc
from sqlalchemy.engine import Result, ScalarResult
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.sql import Executable
from sqlalchemy.sql.selectable import TypedReturnsRows

from notora.v2.exceptions.common import AlreadyExistsError, FKNotFoundError, NotFoundError
from notora.v2.models.base import GenericBaseModel


class SessionExecutorMixin[PKType, ModelType: GenericBaseModel]:
    _unique_violation_errors: ClassVar[dict[str, str]] = {}
    _fk_violation_errors: ClassVar[dict[str, str]] = {}

    _unique_constraint_pattern = re.compile(
        r'.*duplicate key value violates unique constraint "(?P<name>\w+)"',
    )
    _fk_constraint_pattern = re.compile(
        r'.*insert or update on table "(?P<table_name>\w+)" '
        r'violates foreign key constraint "(?P<fk_name>\w+)"',
    )

    @property
    def _not_found_error(self) -> str:
        return f'{self.__class__.__name__.removesuffix("Service")} not found.'

    async def execute(
        self,
        session: AsyncSession,
        statement: Executable,
    ) -> Result[Any]:
        try:
            return await session.execute(statement)
        except exc.IntegrityError as err:
            raise self._translate_integrity_error(err) from err

    async def execute_scalars(
        self,
        session: AsyncSession,
        statement: Executable,
    ) -> ScalarResult[Any]:
        result = await self.execute(session, statement)
        return result.scalars()

    async def execute_scalars_all(
        self,
        session: AsyncSession,
        statement: Executable,
    ) -> list[Any]:
        return list((await self.execute_scalars(session, statement)).all())

    async def execute_scalar_one(
        self,
        session: AsyncSession,
        statement: Executable,
    ) -> Any:
        result = await self.execute(session, statement)
        return result.scalar_one()

    async def execute_for_one(
        self,
        session: AsyncSession,
        statement: TypedReturnsRows[tuple[ModelType]],
    ) -> ModelType:
        result = await self.execute(session, statement)
        entity = result.unique().scalar_one_or_none()
        if entity is None:
            raise NotFoundError[PKType](self._not_found_error)
        return cast(ModelType, entity)

    async def execute_optional(
        self,
        session: AsyncSession,
        statement: TypedReturnsRows[tuple[ModelType]],
    ) -> ModelType | None:
        result = await self.execute(session, statement)
        return cast(ModelType | None, result.unique().scalar_one_or_none())

    def _translate_integrity_error(self, err: exc.IntegrityError) -> Exception:
        if match := self._fk_constraint_pattern.match(err.args[0]):
            fk_name = match.group('fk_name')
            table_name = match.group('table_name')
            return FKNotFoundError(
                self._fk_violation_errors.get(fk_name, 'Related object not found.'),
                fk_name=fk_name,
                table_name=table_name,
            )
        if match := self._unique_constraint_pattern.match(err.args[0]):
            constraint = match.group('name')
            return AlreadyExistsError(
                self._unique_violation_errors.get(constraint, 'Entity already exists.'),
                constraint_name=constraint,
            )
        return err



from typing import Any

from pydantic import BaseModel as PydanticModel

from notora.v2.models.base import GenericBaseModel


class PayloadMixin[ModelType: GenericBaseModel]:
    @staticmethod
    def _dump_payload(
        data: PydanticModel | dict[str, Any],
        *,
        exclude_unset: bool,
    ) -> dict[str, Any]:
        if isinstance(data, dict):
            return dict(data)
        return data.model_dump(exclude_unset=exclude_unset)



from uuid import uuid4

from sqlalchemy import select
from sqlalchemy.ext.asyncio import AsyncSession

from notora.v2.services.mixins.m2m import M2MSyncMode
from tests.v2.test_integration.mocks.model import V2Role, V2UserRole
from tests.v2.test_integration.mocks.schema import V2UserResponseSchema
from tests.v2.test_integration.mocks.service import V2UserService


async def _create_role(session: AsyncSession, name: str) -> V2Role:
    role = V2Role(id=uuid4(), name=name)
    session.add(role)
    await session.commit()
    return role


async def test_m2m_replace_on_update(
    db_session: AsyncSession,
    user_service: V2UserService,
) -> None:
    role_a = await _create_role(db_session, 'role-a')
    role_b = await _create_role(db_session, 'role-b')
    role_c = await _create_role(db_session, 'role-c')

    created = await user_service.create(
        db_session,
        {
            'id': uuid4(),
            'email': 'm2m@ex.com',
            'name': 'M2M',
            'is_active': True,
            'role_ids': [role_a.id, role_b.id],
        },
    )
    await db_session.commit()
    assert isinstance(created, V2UserResponseSchema)

    role_rows = (
        await db_session.scalars(select(V2UserRole).where(V2UserRole.user_id == created.id))
    ).all()
    assert {row.role_id for row in role_rows} == {role_a.id, role_b.id}

    await user_service.update(
        db_session,
        created.id,
        {
            'name': 'M2M Updated',
            'role_ids': [role_c.id],
        },
    )
    await db_session.commit()

    updated_rows = (
        await db_session.scalars(select(V2UserRole).where(V2UserRole.user_id == created.id))
    ).all()
    assert {row.role_id for row in updated_rows} == {role_c.id}


async def test_m2m_add_mode(
    db_session: AsyncSession,
    user_service: V2UserService,
) -> None:
    role_a = await _create_role(db_session, 'add-a')
    role_b = await _create_role(db_session, 'add-b')

    user_service.m2m_sync_mode = M2MSyncMode.ADD
    created = await user_service.create(
        db_session,
        {
            'id': uuid4(),
            'email': 'add@ex.com',
            'name': 'Add',
            'is_active': True,
            'role_ids': [role_a.id],
        },
    )
    await db_session.commit()
    assert isinstance(created, V2UserResponseSchema)

    await user_service.update(
        db_session,
        created.id,
        {
            'name': 'Add Updated',
            'role_ids': [role_a.id, role_b.id],
        },
    )
    await db_session.commit()

    rows = (
        await db_session.scalars(select(V2UserRole).where(V2UserRole.user_id == created.id))
    ).all()
    assert {row.role_id for row in rows} == {role_a.id, role_b.id}



import re
import uuid
from datetime import datetime
from typing import Any

from sqlalchemy import DateTime, ForeignKey, MetaData, func
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    declared_attr,
    mapped_column,
    relationship,
)
from sqlalchemy.sql.functions import GenericFunction

POSTGRES_NAMING_CONVENTION = {
    'ix': '%(column_0_label)s_idx',
    'uq': '%(table_name)s_%(column_0_name)s_key',
    'ck': '%(table_name)s_%(constraint_name)s_check',
    'fk': '%(table_name)s_%(column_0_name)s_fkey',
    'pk': '%(table_name)s_pkey',
}
metadata = MetaData(naming_convention=POSTGRES_NAMING_CONVENTION)


class Base(DeclarativeBase):
    metadata = metadata


class GenerateUUID(GenericFunction[uuid.UUID]):
    type = UUID()
    name = 'gen_random_uuid'
    identifier = 'gen_random_uuid'


class CreatableMixin:
    created_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=False),
        server_default=func.now(),
    )


class UpdatableMixin:
    updated_at: Mapped[datetime] = mapped_column(
        DateTime(timezone=False),
        server_default=func.now(),
        onupdate=func.now(),
    )

    @declared_attr
    @classmethod
    def updated_by(cls) -> Mapped[uuid.UUID | None]:
        return mapped_column(ForeignKey('user.id', use_alter=True))

    @declared_attr
    @classmethod
    def updated_by_user(cls) -> Mapped[uuid.UUID | None]:
        return relationship('User', foreign_keys=[cls.updated_by])


class SoftDeletableMixin:
    deleted_at: Mapped[datetime | None] = mapped_column(
        DateTime(timezone=False),
        nullable=True,
        default=None,
    )


class GenericBaseModel(Base, CreatableMixin):
    __abstract__ = True
    id: Mapped[uuid.UUID] = mapped_column(
        UUID,
        primary_key=True,
        server_default=GenerateUUID(),
    )

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} id={self.id}>'

    def to_dict(self) -> dict[str, Any]:
        return {field.name: getattr(self, field.name) for field in self.__table__.columns}


class BaseModel(GenericBaseModel, UpdatableMixin, SoftDeletableMixin):
    __abstract__ = True

    @declared_attr  # type: ignore[arg-type]
    def __tablename__(cls) -> str:  # noqa: N805
        return re.compile(r'(?<!^)(?=[A-Z])').sub('_', cls.__name__).lower()


class UpdatableModel(GenericBaseModel, UpdatableMixin):
    __abstract__ = True


class SoftDeletableModel(GenericBaseModel, SoftDeletableMixin):
    __abstract__ = True



from uuid import UUID

from notora.v1.persistence.repos.base import SoftDeletableRepo
from tests.v1.test_integration.mocks.model import MockModel


class MockRepo(SoftDeletableRepo[UUID, MockModel]): ...



from sqlalchemy import String, select
from sqlalchemy.orm import Mapped, load_only, mapped_column

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.base import Repository
from notora.v2.repositories.config import RepoConfig
from notora.v2.repositories.params import QueryParams


class Widget(GenericBaseModel):
    __tablename__ = 'widget'

    name: Mapped[str] = mapped_column(String)


def test_list_uses_default_limit() -> None:
    repo = Repository[object, Widget](Widget)
    stmt = repo.list()
    assert ' LIMIT ' in str(stmt)


def test_list_supports_no_limit() -> None:
    repo = Repository[object, Widget](Widget)
    stmt = repo.list(limit=None)
    assert ' LIMIT ' not in str(stmt)


def test_list_applies_offset_without_limit() -> None:
    repo = Repository[object, Widget](Widget)
    stmt = repo.list(limit=None, offset=10)
    compiled = str(stmt)
    assert ' LIMIT ' not in compiled or ' LIMIT -1 ' in compiled
    assert ' OFFSET ' in compiled


def test_list_applies_options_to_base_query() -> None:
    repo = Repository[object, Widget](Widget)
    base_query = select(Widget)
    option = load_only(Widget.name)
    stmt = repo.list(base_query=base_query, options=(option,))
    assert stmt._with_options


def test_repo_config_overrides_defaults() -> None:
    default_limit = 5
    config = RepoConfig[Widget](default_limit=default_limit, pk_attribute='id')
    repo = Repository[object, Widget](Widget, config=config)
    assert repo.default_limit == default_limit
    assert repo.pk_attribute == 'id'


def test_list_by_params_respects_limit_override() -> None:
    repo = Repository[object, Widget](Widget)
    params = QueryParams[Widget](limit=None)
    stmt = repo.list_by_params(params)
    assert ' LIMIT ' not in str(stmt)


def test_list_uses_fallback_sort_attribute() -> None:
    repo = Repository[object, Widget](Widget)
    stmt = repo.list(limit=None)
    assert ' ORDER BY ' in str(stmt)
    assert 'widget.id' in str(stmt)


def test_repo_config_customizes_fallback_sort_attribute() -> None:
    config = RepoConfig[Widget](fallback_sort_attribute='name')
    repo = Repository[object, Widget](Widget, config=config)
    stmt = repo.list(limit=None)
    compiled = str(stmt)
    assert ' ORDER BY ' in compiled
    assert 'widget.name' in compiled



from collections.abc import Sequence
from typing import Annotated

from pydantic import BaseModel, ConfigDict, PlainSerializer

from notora.types import AnyIPAddress


class BaseResponseSchema(BaseModel):
    model_config = ConfigDict(from_attributes=True)


class BaseRequestSchema(BaseModel):
    model_config = ConfigDict(from_attributes=True)


class PaginationMetaSchema(BaseModel):
    limit: int
    offset: int
    total: int

    @classmethod
    def calculate(cls, total: int, limit: int, offset: int) -> 'PaginationMetaSchema':
        if limit <= 0:
            msg = 'limit must be a positive integer.'
            raise ValueError(msg)
        if offset < 0:
            msg = 'offset must be zero or a positive integer.'
            raise ValueError(msg)
        total_value = max(total, 0)
        return cls(
            limit=limit,
            offset=offset,
            total=total_value,
        )


class PaginatedResponseSchema[T](BaseResponseSchema):
    meta: PaginationMetaSchema
    data: Sequence[T]


class ClientMeta(BaseRequestSchema):
    ip_address: Annotated[AnyIPAddress, PlainSerializer(str, return_type=str)] | None = None
    user_agent: str | None = None



from collections.abc import Iterable, Sequence
from typing import Any

from pydantic import BaseModel as PydanticModel
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import InstrumentedAttribute

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.types import FilterSpec, OptionSpec
from notora.v2.schemas.base import BaseResponseSchema
from notora.v2.services.mixins.accessors import RepositoryAccessorMixin
from notora.v2.services.mixins.executor import SessionExecutorMixin
from notora.v2.services.mixins.m2m import ManyToManySyncMixin
from notora.v2.services.mixins.payload import PayloadMixin
from notora.v2.services.mixins.serializer import SerializerProtocol
from notora.v2.services.mixins.updated_by import UpdatedByServiceMixin

__all__ = ['CreateOrSkipServiceMixin', 'CreateServiceMixin']


class CreateServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    ManyToManySyncMixin[PKType, ModelType],
    UpdatedByServiceMixin[PKType, ModelType],
    PayloadMixin[ModelType],
    SerializerProtocol[ModelType, DetailSchema, ListSchema],
):
    async def create_raw(
        self,
        session: AsyncSession,
        data: PydanticModel | dict[str, Any],
        *,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ModelType:
        payload = self._dump_payload(data, exclude_unset=False)
        payload, relation_payload = self.split_m2m_payload(payload)
        payload = self._apply_updated_by(payload, actor_id)
        query = self.repo.create(payload, options=options)
        entity = await self.execute_for_one(session, query)
        if relation_payload:
            await self.sync_m2m_relations(session, self._extract_pk(entity), relation_payload)
        return entity

    async def create(
        self,
        session: AsyncSession,
        data: PydanticModel | dict[str, Any],
        *,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: type[DetailSchema] | None = None,
    ) -> DetailSchema:
        entity = await self.create_raw(session, data, actor_id=actor_id, options=options)
        return self.serialize_one(entity, schema=schema)


class CreateOrSkipServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    SessionExecutorMixin[PKType, ModelType],
    RepositoryAccessorMixin[PKType, ModelType],
    UpdatedByServiceMixin[PKType, ModelType],
    PayloadMixin[ModelType],
    SerializerProtocol[ModelType, DetailSchema, ListSchema],
):
    async def create_or_skip_raw(
        self,
        session: AsyncSession,
        data: PydanticModel | dict[str, Any],
        *,
        conflict_columns: Sequence[InstrumentedAttribute[Any]],
        conflict_where: Iterable[FilterSpec[ModelType]] | None = None,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ModelType | None:
        payload = self._dump_payload(data, exclude_unset=False)
        payload = self._apply_updated_by(payload, actor_id)
        query = self.repo.create_or_skip(
            payload,
            conflict_columns=conflict_columns,
            conflict_where=conflict_where,
            options=options,
        )
        return await self.execute_optional(session, query)

    async def create_or_skip(
        self,
        session: AsyncSession,
        data: PydanticModel | dict[str, Any],
        *,
        conflict_columns: Sequence[InstrumentedAttribute[Any]],
        conflict_where: Iterable[FilterSpec[ModelType]] | None = None,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: type[DetailSchema] | None = None,
    ) -> DetailSchema | None:
        entity = await self.create_or_skip_raw(
            session,
            data,
            conflict_columns=conflict_columns,
            conflict_where=conflict_where,
            actor_id=actor_id,
            options=options,
        )
        if entity is None:
            return None
        return self.serialize_one(entity, schema=schema)



from collections.abc import Iterable
from typing import overload

from sqlalchemy.ext.asyncio import AsyncSession

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.types import FilterSpec, OptionSpec, OrderSpec
from notora.v2.schemas.base import BaseResponseSchema
from notora.v2.services.mixins.accessors import RepositoryAccessorMixin
from notora.v2.services.mixins.executor import SessionExecutorMixin
from notora.v2.services.mixins.serializer import SerializerProtocol

__all__ = ['RetrievalServiceMixin']

class RetrievalServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    SessionExecutorMixin[PKType, ModelType],
    RepositoryAccessorMixin[PKType, ModelType],
    SerializerProtocol[ModelType, DetailSchema, ListSchema],
):
    __type_params__: tuple[object, ...]

    async def retrieve_raw(
        self,
        session: AsyncSession,
        pk: PKType,
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ModelType: ...
    async def retrieve_one_by_raw(
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ModelType: ...
    @overload
    async def retrieve(
        self,
        session: AsyncSession,
        pk: PKType,
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: None = ...,
    ) -> DetailSchema: ...
    @overload
    async def retrieve[SchemaT: BaseResponseSchema](
        self,
        session: AsyncSession,
        pk: PKType,
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: type[SchemaT],
    ) -> SchemaT: ...
    @overload
    async def retrieve_one_by(
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: None = ...,
    ) -> DetailSchema: ...
    @overload
    async def retrieve_one_by[SchemaT: BaseResponseSchema](
        self,
        session: AsyncSession,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: type[SchemaT],
    ) -> SchemaT: ...



from typing import cast

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.base import Repository, SoftDeleteRepository
from notora.v2.repositories.config import RepoConfig

type AnyRepository[
    PKType,
    ModelType: GenericBaseModel,
] = Repository[PKType, ModelType] | SoftDeleteRepository[PKType, ModelType]

type RepositoryType[
    PKType,
    ModelType: GenericBaseModel,
] = type[Repository[PKType, ModelType]] | type[SoftDeleteRepository[PKType, ModelType]]


def build_repository[PKType, ModelType: GenericBaseModel](
    model: type[ModelType],
    *,
    config: RepoConfig[ModelType] | None = None,
    soft_delete: bool = False,
    repo_cls: RepositoryType[PKType, ModelType] | None = None,
) -> AnyRepository[PKType, ModelType]:
    """Create a repository with optional config overrides."""
    if repo_cls is None:
        repo_cls = cast(
            RepositoryType[PKType, ModelType],
            SoftDeleteRepository if soft_delete else Repository,
        )
    return repo_cls(model, config=config)



from collections.abc import Iterable, Sequence
from typing import Any, cast

from sqlalchemy import and_, delete, func, insert, update
from sqlalchemy.dialects.postgresql import insert as pg_insert
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.sql.dml import ReturningInsert
from sqlalchemy.sql.selectable import TypedReturnsRows

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.mixins.query import FilterableMixin, LoadOptionsMixin, PrimaryKeyMixin
from notora.v2.repositories.types import FilterClause, FilterSpec, OptionSpec


class CreateMixin[ModelType: GenericBaseModel](LoadOptionsMixin[ModelType]):
    def create(
        self,
        payload: dict[str, Any],
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ReturningInsert[tuple[ModelType]]:
        stmt = insert(self.model).values(**payload).returning(self.model)
        return self.apply_options(stmt, options)

    def bulk_create(
        self,
        payload: Sequence[dict[str, Any]],
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ReturningInsert[tuple[ModelType]]:
        stmt = insert(self.model).values(list(payload)).returning(self.model)
        return self.apply_options(stmt, options)


class UpsertMixin[PKType, ModelType: GenericBaseModel](
    PrimaryKeyMixin[PKType, ModelType],
    LoadOptionsMixin[ModelType],
    FilterableMixin[ModelType],
):
    def upsert(
        self,
        payload: dict[str, Any],
        *,
        conflict_columns: Sequence[InstrumentedAttribute[Any]] | None = None,
        conflict_where: Iterable[FilterSpec[ModelType]] | None = None,
        update_only: Sequence[str] | None = None,
        update_exclude: Sequence[str] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ReturningInsert[tuple[ModelType]]:
        if update_only is not None and update_exclude is not None:
            msg = 'Only one of update_only or update_exclude can be provided.'
            raise ValueError(msg)
        stmt = pg_insert(self.model).values(**payload)

        if update_only is not None:
            update_payload = {k: v for k, v in payload.items() if k in update_only}
        elif update_exclude is not None:
            update_payload = {k: v for k, v in payload.items() if k not in update_exclude}
        else:
            update_payload = payload

        clauses = self.merge_filters(conflict_where)
        where_clause = and_(*clauses) if clauses else None

        index_elements = conflict_columns or (cast(InstrumentedAttribute[Any], self.pk_column),)
        returning_stmt = stmt.on_conflict_do_update(
            index_elements=cast(Iterable[Any], index_elements),
            index_where=where_clause,
            set_=update_payload,
        ).returning(self.model)
        return self.apply_options(returning_stmt, options)


class CreateOrSkipMixin[ModelType: GenericBaseModel](
    LoadOptionsMixin[ModelType],
    FilterableMixin[ModelType],
):
    def create_or_skip(
        self,
        payload: dict[str, Any],
        *,
        conflict_columns: Sequence[InstrumentedAttribute[Any]],
        conflict_where: Iterable[FilterSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ReturningInsert[tuple[ModelType]]:
        stmt = pg_insert(self.model).values(**payload)
        clauses = self.merge_filters(conflict_where)
        where_clause = and_(*clauses) if clauses else None
        returning_stmt = stmt.on_conflict_do_nothing(
            index_elements=cast(Iterable[Any], conflict_columns),
            index_where=where_clause,
        ).returning(self.model)
        return self.apply_options(returning_stmt, options)


class UpdateMixin[PKType, ModelType: GenericBaseModel](
    PrimaryKeyMixin[PKType, ModelType],
    LoadOptionsMixin[ModelType],
    FilterableMixin[ModelType],
):
    def update_by(
        self,
        payload: dict[str, Any],
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]:
        stmt = update(self.model).values(**payload)
        stmt = self.apply_filters(stmt, filters)
        stmt = stmt.returning(self.model)
        return self.apply_options(stmt, options)

    def update(
        self,
        pk: PKType,
        payload: dict[str, Any],
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]:
        filters = (cast(FilterClause, self.pk_column == pk),)
        return self.update_by(payload, filters=filters, options=options)


class DeleteMixin[PKType, ModelType: GenericBaseModel](
    PrimaryKeyMixin[PKType, ModelType],
    LoadOptionsMixin[ModelType],
    FilterableMixin[ModelType],
):
    def delete_by(
        self,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]:
        stmt = delete(self.model)
        stmt = self.apply_filters(stmt, filters)
        returning_stmt = stmt.returning(self.model)
        return self.apply_options(returning_stmt, options)

    def delete(
        self,
        pk: PKType,
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]:
        filters = (cast(FilterClause, self.pk_column == pk),)
        return self.delete_by(filters=filters, options=options)


class SoftDeleteMixin[PKType, ModelType: GenericBaseModel](UpdateMixin[PKType, ModelType]):
    deleted_attribute: str = 'deleted_at'

    def soft_delete_by(
        self,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]:
        payload = {self.deleted_attribute: func.now()}
        return super().update_by(payload, filters=filters, options=options)

    def soft_delete(
        self,
        pk: PKType,
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> TypedReturnsRows[tuple[ModelType]]:
        filters = (cast(FilterClause, self.pk_column == pk),)
        return self.soft_delete_by(filters=filters, options=options)



from typing import Any

from faker import Faker
from polyfactory import Ignore
from polyfactory.factories.pydantic_factory import ModelFactory
from polyfactory.factories.sqlalchemy_factory import SQLAlchemyFactory
from pydantic import BaseModel as PydanticBaseModel

from notora.v2.models.base import BaseModel as SQLAlchemyBaseModel
from notora.v2.schemas.base import BaseRequestSchema, BaseResponseSchema

_f = Faker()


class SQLAlchemyBaseModelFactory[T: SQLAlchemyBaseModel](SQLAlchemyFactory[T]):
    __faker__ = _f
    __is_base_factory__ = True
    __set_foreign_keys__ = False
    __set_relationships__ = False
    __check_model__ = False

    deleted_at = Ignore()
    updated_by = Ignore()
    updated_by_user = None


class BaseModelFactory[T: PydanticBaseModel](ModelFactory[T]):
    __faker__ = _f
    __is_base_factory__ = True


class BaseRequestFactory[T: BaseRequestSchema](BaseModelFactory[T]):
    __is_base_factory__ = True

    @classmethod
    def build_json(cls, **kwargs: Any) -> dict[str, Any]:
        return cls.build(**kwargs).model_dump(mode='json')


class BaseResponseFactory[T: BaseResponseSchema](BaseModelFactory[T]):
    __is_base_factory__ = True

    @classmethod
    def build_json(cls, **kwargs: Any) -> dict[str, Any]:
        return cls.build(**kwargs).model_dump(mode='json')



from collections.abc import Iterable, Sequence
from typing import Any

from pydantic import BaseModel as PydanticModel
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import InstrumentedAttribute

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.types import FilterSpec, OptionSpec
from notora.v2.schemas.base import BaseResponseSchema
from notora.v2.services.mixins.m2m import ManyToManySyncMixin
from notora.v2.services.mixins.payload import PayloadMixin
from notora.v2.services.mixins.serializer import SerializerProtocol
from notora.v2.services.mixins.updated_by import UpdatedByServiceMixin

__all__ = ['UpsertServiceMixin']


class UpsertServiceMixin[
    PKType,
    ModelType: GenericBaseModel,
    DetailSchema: BaseResponseSchema,
    ListSchema: BaseResponseSchema = DetailSchema,
](
    ManyToManySyncMixin[PKType, ModelType],
    UpdatedByServiceMixin[PKType, ModelType],
    PayloadMixin[ModelType],
    SerializerProtocol[ModelType, DetailSchema, ListSchema],
):
    async def upsert_raw(
        self,
        session: AsyncSession,
        data: PydanticModel | dict[str, Any],
        *,
        conflict_columns: Sequence[InstrumentedAttribute[Any]] | None = None,
        conflict_where: Iterable[FilterSpec[ModelType]] | None = None,
        update_only: Sequence[str] | None = None,
        update_exclude: Sequence[str] | None = None,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> ModelType:
        payload = self._dump_payload(data, exclude_unset=False)
        payload, relation_payload = self.split_m2m_payload(payload)
        payload = self._apply_updated_by(payload, actor_id)
        query = self.repo.upsert(
            payload,
            conflict_columns=conflict_columns,
            conflict_where=conflict_where,
            update_only=update_only,
            update_exclude=update_exclude,
            options=options,
        )
        entity = await self.execute_for_one(session, query)
        if relation_payload:
            await self.sync_m2m_relations(session, self._extract_pk(entity), relation_payload)
        return entity

    async def upsert(
        self,
        session: AsyncSession,
        data: PydanticModel | dict[str, Any],
        *,
        conflict_columns: Sequence[InstrumentedAttribute[Any]] | None = None,
        conflict_where: Iterable[FilterSpec[ModelType]] | None = None,
        update_only: Sequence[str] | None = None,
        update_exclude: Sequence[str] | None = None,
        actor_id: Any | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        schema: type[DetailSchema] | None = None,
    ) -> DetailSchema:
        entity = await self.upsert_raw(
            session,
            data,
            conflict_columns=conflict_columns,
            conflict_where=conflict_where,
            update_only=update_only,
            update_exclude=update_exclude,
            actor_id=actor_id,
            options=options,
        )
        return self.serialize_one(entity, schema=schema)



from collections.abc import Iterable, Sequence
from typing import Any, cast

from sqlalchemy import Select, func, select
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.sql.base import ExecutableOption

from notora.v2.models.base import GenericBaseModel
from notora.v2.repositories.params import QueryParams
from notora.v2.repositories.types import (
    DEFAULT_LIMIT,
    DefaultLimit,
    FilterClause,
    FilterSpec,
    OptionSpec,
    OrderClause,
    OrderSpec,
    SupportsOptions,
    SupportsWhere,
)


class LoadOptionsMixin[ModelType: GenericBaseModel]:
    """Adds ``.options`` support with overridable defaults."""

    model: type[ModelType]
    default_options: Sequence[OptionSpec[ModelType]] = ()

    def _resolve_option(self, spec: OptionSpec[ModelType]) -> ExecutableOption:
        return spec(self.model) if callable(spec) else spec

    def merge_options(
        self,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> tuple[ExecutableOption, ...]:
        custom = tuple(options or ())
        specs = (*self.default_options, *custom)
        return tuple(self._resolve_option(spec) for spec in specs)

    def apply_options[StatementT: SupportsOptions](
        self,
        statement: StatementT,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> StatementT:
        merged = self.merge_options(options)
        if merged:
            statement = statement.options(*merged)
        return statement


class FilterableMixin[ModelType: GenericBaseModel]:
    model: type[ModelType]
    default_filters: Sequence[FilterSpec[ModelType]] = ()

    def _resolve_filter(self, spec: FilterSpec[ModelType]) -> FilterClause:
        return spec(self.model) if callable(spec) else spec

    def merge_filters(
        self,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
    ) -> tuple[FilterClause, ...]:
        custom = tuple(filters or ())
        specs = (*self.default_filters, *custom)
        return tuple(self._resolve_filter(spec) for spec in specs)

    def apply_filters[StatementT: SupportsWhere](
        self,
        statement: StatementT,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
    ) -> StatementT:
        for clause in self.merge_filters(filters):
            statement = statement.where(clause)
        return statement


class OrderableMixin[ModelType: GenericBaseModel]:
    model: type[ModelType]
    default_ordering: Sequence[OrderSpec[ModelType]] = ()
    fallback_sort_attribute: str = 'id'

    def _resolve_order(self, spec: OrderSpec[ModelType]) -> OrderClause:
        return spec(self.model) if callable(spec) else spec

    def merge_ordering(
        self,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
    ) -> tuple[Any, ...]:
        custom = tuple(ordering or ())
        specs = (*self.default_ordering, *custom)
        resolved = [self._resolve_order(spec) for spec in specs]
        if not resolved and hasattr(self.model, self.fallback_sort_attribute):
            pk_column = getattr(self.model, self.fallback_sort_attribute)
            resolved.append(pk_column.asc())
        return tuple(resolved)

    def apply_ordering(
        self,
        statement: Select[Any],
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
    ) -> Select[Any]:
        clauses = self.merge_ordering(ordering)
        if clauses:
            statement = statement.order_by(*clauses)
        return statement


class SelectableMixin[ModelType: GenericBaseModel](LoadOptionsMixin[ModelType]):
    def select(
        self,
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> Select[tuple[ModelType]]:
        stmt = select(self.model)
        return self.apply_options(stmt, options)


class ListableMixin[ModelType: GenericBaseModel](
    SelectableMixin[ModelType],
    FilterableMixin[ModelType],
    OrderableMixin[ModelType],
):
    default_limit: int = 50

    def list(
        self,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        limit: int | DefaultLimit | None = DEFAULT_LIMIT,
        offset: int = 0,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
        base_query: Select[tuple[ModelType]] | None = None,
    ) -> Select[tuple[ModelType]]:
        if base_query is None:
            stmt = self.select(options=options)
        else:
            stmt = self.apply_options(base_query, options)
        stmt = self.apply_filters(stmt, filters)
        stmt = self.apply_ordering(stmt, ordering)
        if limit is DEFAULT_LIMIT:
            limit_value = self.default_limit
        elif limit is None:
            limit_value = None
        else:
            limit_value = cast(int | None, limit)
        if limit_value is not None:
            stmt = stmt.limit(limit_value)
        if offset:
            stmt = stmt.offset(offset)
        return stmt

    def list_by_params(self, params: QueryParams[ModelType]) -> Select[tuple[ModelType]]:
        return self.list(
            filters=params.filters,
            limit=params.limit,
            offset=params.offset,
            ordering=params.ordering,
            options=params.options,
            base_query=params.base_query,
        )


class PrimaryKeyMixin[PKType, ModelType: GenericBaseModel]:
    model: type[ModelType]
    pk_attribute: str = 'id'

    @property
    def pk_column(self) -> InstrumentedAttribute[PKType]:
        return cast(InstrumentedAttribute[PKType], getattr(self.model, self.pk_attribute))


class RetrievableMixin[PKType, ModelType: GenericBaseModel](
    PrimaryKeyMixin[PKType, ModelType],
    ListableMixin[ModelType],
):
    def retrieve(
        self,
        pk: PKType,
        *,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> Select[tuple[ModelType]]:
        return self.list(
            filters=(cast(FilterClause, self.pk_column == pk),),
            limit=1,
            options=options,
        )

    def retrieve_by(
        self,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> Select[tuple[ModelType]]:
        stmt = self.select(options=options)
        stmt = self.apply_filters(stmt, filters)
        stmt = self.apply_ordering(stmt, ordering)
        return stmt

    def retrieve_one_by(
        self,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
        ordering: Iterable[OrderSpec[ModelType]] | None = None,
        options: Iterable[OptionSpec[ModelType]] | None = None,
    ) -> Select[tuple[ModelType]]:
        stmt = self.retrieve_by(filters=filters, ordering=ordering, options=options)
        return stmt.limit(1)


class CountableMixin[ModelType: GenericBaseModel](FilterableMixin[ModelType]):
    def count(
        self,
        *,
        filters: Iterable[FilterSpec[ModelType]] | None = None,
    ) -> Select[tuple[int]]:
        stmt = select(func.count()).select_from(self.model)
        return self.apply_filters(stmt, filters)



from uuid import uuid4

import pytest
from sqlalchemy.ext.asyncio import AsyncSession

from notora.v2.exceptions.common import NotFoundError
from notora.v2.repositories.params import PaginationParams, QueryParams
from notora.v2.schemas.base import PaginatedResponseSchema
from tests.v2.test_integration.mocks.model import V2User
from tests.v2.test_integration.mocks.schema import V2UserCreateSchema, V2UserResponseSchema
from tests.v2.test_integration.mocks.service import V2UserService


def _create_user_payload(email: str, name: str) -> V2UserCreateSchema:
    return V2UserCreateSchema(id=uuid4(), email=email, name=name, is_active=True)


async def test_service_create_update_and_updated_by(
    db_session: AsyncSession,
    user_service: V2UserService,
) -> None:
    actor_id = uuid4()
    payload = _create_user_payload('service@ex.com', 'Service')

    created = await user_service.create(db_session, payload, actor_id=actor_id)
    await db_session.commit()

    assert isinstance(created, V2UserResponseSchema)
    assert created.updated_by == actor_id

    new_actor_id = uuid4()
    updated = await user_service.update(
        db_session,
        created.id,
        {'name': 'Service Updated'},
        actor_id=new_actor_id,
    )
    await db_session.commit()

    assert isinstance(updated, V2UserResponseSchema)
    assert updated.updated_by == new_actor_id

    refreshed = await db_session.get(V2User, created.id)
    assert refreshed is not None
    assert refreshed.updated_by == new_actor_id


async def test_service_update_by_and_soft_delete(
    db_session: AsyncSession,
    user_service: V2UserService,
) -> None:
    payload = _create_user_payload('update-by@ex.com', 'Before')
    created = await user_service.create(db_session, payload)
    await db_session.commit()

    assert isinstance(created, V2UserResponseSchema)
    updated = await user_service.update_by(
        db_session,
        filters=[V2User.email == payload.email],
        data={'name': 'After'},
    )
    await db_session.commit()

    assert isinstance(updated, V2UserResponseSchema)
    assert updated.name == 'After'

    await user_service.soft_delete(db_session, created.id)
    await db_session.commit()

    refreshed = await db_session.get(V2User, created.id)
    assert refreshed is not None
    assert refreshed.deleted_at is not None


async def test_service_list_and_paginate_params(
    db_session: AsyncSession,
    user_service: V2UserService,
) -> None:
    payloads = [
        _create_user_payload('a@ex.com', 'A'),
        _create_user_payload('b@ex.com', 'B'),
        _create_user_payload('c@ex.com', 'C'),
    ]
    for payload in payloads:
        await user_service.create(db_session, payload)
    await db_session.commit()

    params = QueryParams[V2User](
        filters=[V2User.email != 'b@ex.com'],
        ordering=[V2User.email.asc()],
        limit=None,
    )
    items = await user_service.list_params(db_session, params)

    assert isinstance(items[0], V2UserResponseSchema)
    assert [item.email for item in items] == ['a@ex.com', 'c@ex.com']

    limit = 2
    page = await user_service.paginate(
        db_session,
        limit=limit,
        offset=0,
    )
    assert isinstance(page, PaginatedResponseSchema)
    assert page.meta.total == len(payloads)
    assert len(page.data) == limit

    limit_param = 1
    page_params = await user_service.paginate_params(
        db_session,
        PaginationParams[V2User](limit=limit_param, offset=1),
    )
    assert page_params.meta.offset == limit_param


async def test_service_retrieve_create_or_skip_upsert_and_delete(
    db_session: AsyncSession,
    user_service: V2UserService,
) -> None:
    payload = _create_user_payload('retrieve@ex.com', 'Retrieve')
    created = await user_service.create(db_session, payload)
    await db_session.commit()

    assert isinstance(created, V2UserResponseSchema)

    retrieved = await user_service.retrieve(db_session, created.id)
    assert isinstance(retrieved, V2UserResponseSchema)

    retrieved_by = await user_service.retrieve_one_by(
        db_session,
        filters=[V2User.email == payload.email],
    )
    assert isinstance(retrieved_by, V2UserResponseSchema)

    created_or_skip = await user_service.create_or_skip(
        db_session,
        {
            'id': uuid4(),
            'email': 'skip@ex.com',
            'name': 'Skip',
            'is_active': True,
        },
        conflict_columns=[V2User.email],
    )
    await db_session.commit()
    assert isinstance(created_or_skip, V2UserResponseSchema)

    duplicate = await user_service.create_or_skip(
        db_session,
        {
            'id': uuid4(),
            'email': created_or_skip.email,
            'name': 'Duplicate',
            'is_active': True,
        },
        conflict_columns=[V2User.email],
    )
    await db_session.commit()
    assert duplicate is None

    upserted = await user_service.upsert(
        db_session,
        {
            'id': uuid4(),
            'email': payload.email,
            'name': 'Upserted',
            'is_active': True,
        },
        conflict_columns=[V2User.email],
        update_only=['name'],
    )
    await db_session.commit()
    assert isinstance(upserted, V2UserResponseSchema)

    refreshed = await db_session.get(V2User, created.id)
    assert refreshed is not None
    assert refreshed.name == 'Upserted'

    await user_service.delete(db_session, created.id)
    await db_session.commit()
    assert await db_session.get(V2User, created.id) is None

    with pytest.raises(NotFoundError):
        await user_service.retrieve(db_session, created.id)
